 
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it),	 *
 *                  Luigi Malagò (malago@elet.polimi.it)                         *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>. 		 *	
 *********************************************************************************/

#include "BinaryInstance.h"
#include <stdlib.h>

/* Default constructor, does not initialize the individual vector. */
BinaryInstance::BinaryInstance() {
	_parent = false; 
	_validCache = false; 
}

/* Constructor, initializes the individual vector to a false vector (all 0s). */
BinaryInstance::BinaryInstance(int size) : _size(size) {
	_individual.insert(_individual.begin(), _size, false);
	_parent = false;
	_validCache = false;
}

/* Copy constructor. */
BinaryInstance::BinaryInstance(const BinaryInstance &bi) {
	unsigned int i; 
	
	_size = bi._size;
	for (i = 0; i < _size; i++) _individual.push_back(bi._individual[i]);
	_parent = bi._parent;
	_validCache = bi._validCache;
	_fitnessCache = bi._fitnessCache;
}

/* Destructor. */
BinaryInstance::~BinaryInstance() {
	_individual.clear();
}

/* Compares two binary instance and return true if they represent the same individual. */
bool BinaryInstance::compare(BinaryInstance *i1, BinaryInstance *i2) {
	if(i1->getSize() != i2->getSize()) return false;
	for(int i=0; i<i1->getSize(); i++) {
		if(i1->get(i) != i2->get(i)) return false;
	}
	return true;
}

/* Sets the fitness cache for the instance. */
void BinaryInstance::setFitnessCache(double fitness) {
	_fitnessCache = fitness;
	_validCache = true;
}

/* Returns the fitness cache values. */
double BinaryInstance::getFitnessCache() {
	return _fitnessCache;
}

/* Tests if the fitness cache value is valid. */
bool BinaryInstance::validCache() {
	return _validCache;
}

/* Returns the individual size. */
int BinaryInstance::getSize() {
	return _size;
}

/* Returns the value at position i as an integer value. */
int BinaryInstance::get(unsigned int i) {
	int rval = 0;
	if (i >= _size || i < 0) cerr << "[BinaryInstance::get] index (" << i << ") out of bound" << endl;
	rval = _individual[i] ? 1 : 0;
	
	return rval;
}

/* Returns the value at position i following the harmonic encoding {-1, 1}*/
int BinaryInstance::getHarmonicEncoding(unsigned int i) {
	int rval = -1;
	if (i >= _size || i < 0) cerr << "[BinaryInstance::get] index (" << i << ") out of bound" << endl;
	rval = _individual[i] ? 1 : -1;
	
	return rval;
}

/* Sets the value at position i as the integer argument value. */
void BinaryInstance::set(unsigned int i, int value) {
	if (i >= _size || i < 0) cerr << "[BinaryInstance::get] index (" << i << ") out of bound" << endl;
	_individual[i] = value == 0 ? false : true;
	_validCache = false;
}

/* Sets all the position of the individual to a value specified by i. */
void BinaryInstance::setAllBits(int i) {
	unsigned int j; 
	bool val = i == 0 ? false : true;
	
	_validCache = false; 
	for (j = 0; j < _size; j++) _individual[i] = val;  
}

/* Returns a string representation of the individual. */
string BinaryInstance::toString() {
	string rval;
	unsigned int i;
	
	for (i = 0; i < _size; i++) {
		if (_individual[i]) rval.append("1");
		else rval.append("0"); 
	}
	
	return rval;
}

/* Returns an array of char as representation of the individual. The returned array is not a c-string, 
 * so it does not end with \0. Array returned follow the black&white image encoding {0,255} */
char* BinaryInstance::toCharArray() {
	unsigned int i;
	char *rval  = new char[_size];
	
	for (i = 0; i < _size; i++) {
		if (_individual[i]) rval[i] = (char) 255;
		else rval[i] = (char) 0; 	
	}
	
	return rval;
}
