
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/

#include "GraphicsTool.h"

/* Converts an image to a BinaryInstance given its filename complete of its path. The image
 * should be a square image, for algorithm's behaviour reasons. */
BinaryInstance* GraphicsTool::imageToBinaryInstance(char *filename) {
	bool debug = false;
	IplImage* img = 0;
	int i, j, count = 0, val = 0;
	int height, width, size, channels, step;
	uchar *data;
	BinaryInstance *loaded;
	
	/* Loading image file. */
	img = cvLoadImage(filename, CV_LOAD_IMAGE_GRAYSCALE);
	if (!img) {
		cerr << "GraphicsTool::toBinaryInstance Error while loading image file" << endl;
	}
	
	/* Checking image properties. */
	height = img->height;
	width = img->width;
	channels = img->nChannels;
	step = img->widthStep;
	data = (uchar*) img->imageData;
	
	if (debug) {
		clog << "Image properties: " << endl;
		clog << "Channels " << channels << endl;
		clog << "depth " << img->depth << endl;
		clog << "Height " << height << endl;
		clog << "Width " << width << endl;
		clog << "Step" << step << endl;
	}
	
	if (height != width)
		cerr << "GraphicsTool::toBinaryInstance Warning, loaded image is not squared!" << endl;
	
	/* Converting to binary instance. */
	size = height * width;
	loaded = new BinaryInstance(size);
	for (i = 0; i < height; i++) {
		for (j = 0; j < width; j++) {
			if(data[i * step + j] == 0) val = 0;
			if(data[i * step + j] == 255) val = 1;
			loaded->set(count, val);
			count++;
		}
	}
	
	if (debug) clog << "Image loaded and converted!" << endl;
	
	cvReleaseImage(&img);
	return loaded;
}

/* Creates an image corresponding to a visual representation of the passed BinaryInstance, 
 * and store it to the specified destination file. Returns true if the image is created 
 * false otherwise. Instance size must be a perfect square number. */
bool GraphicsTool::binaryInstanceToImage(BinaryInstance *instance, char *destinationFile) {
	unsigned int i, j, k = 0, hw, step;
	double tmp;
	IplImage *img = 0;
	
	tmp = sqrt(instance->getSize());
	hw = (unsigned int) sqrt(instance->getSize());
	if (ceil(tmp) > hw) {
		cerr << "[GraphicsTool::binaryInstanceToImage] Image size is not a perfect square." << endl;
		return false;
	}
	
	img = cvCreateImage(cvSize(hw,hw), IPL_DEPTH_8U, 1);
	step = img->widthStep;
	
	for (i = 0; i < hw; i++) {
		for (j = 0; j < hw; j++) {
			if (instance->get(k) == 0) img->imageData[i * step + j] = 0;
			else img->imageData[i * step + j] = 255;
			k++;
		}
	}
	
	if(!cvSaveImage(destinationFile, img)) {
		cerr << "[GraphicsTool::binaryInstanceToImage] Error while saving image." << endl;
		return false;
	}
	
	cvReleaseImage(&img);
	return true;
}

/* Creates an image corresponding to a visual representation of the passed char array, 
 * and store it to the specified destination file. Returns true if the image is created 
 * false otherwise. Array length must be a perfect square number. */
bool GraphicsTool::binaryInstanceToImage(char *instance, unsigned int length, char *destinationFile) {
	int hw, i, j, k = 0, step;
	double tmp;
	IplImage *img = 0;
	
	tmp = sqrt(length);
	hw = (unsigned int) sqrt(length);
	if (ceil(tmp) > hw) {
		cerr << "[GraphicsTool::binaryInstanceToImage] Image size is not a perfect square." << endl;
		return false;
	}
	
	img = cvCreateImage(cvSize(hw,hw), IPL_DEPTH_8U, 1);
	step = img->widthStep;
	
	for (i = 0; i < hw; i++) {
		for (j = 0; j < hw; j++) {
			img->imageData[i * step + j] = instance[k];
			k++;
		}
	}

	if(!cvSaveImage(destinationFile, img)) {
		cerr << "[GraphicsTool::binaryInstanceToImage] Error while saving image." << endl;
		return false;
	}
	
	cvReleaseImage(&img);
	return true;
}
