
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/

#include <PermutationInstance.h>

/* Create a new permutation instance without initialize it */
PermutationInstance::PermutationInstance(int size) {
	instance_ = gsl_permutation_alloc(size);
	valid_cache_ = false; 
}

/* Create a new permutation instance. If randomized is true then the permutation is a random one, 
   otherwise it is initialized to the identity (0, 1, .. size - 1)*/
PermutationInstance::PermutationInstance(int size, bool randomized, const gsl_rng *r) {
	if(randomized) {
		instance_ = gsl_permutation_calloc(size);
     		gsl_ran_shuffle (r, instance_->data, size, sizeof(size_t));
     		valid_cache_ = false;
	} else {
		instance_ = gsl_permutation_calloc(size);
		valid_cache_ = false;
	}
}

/* Create a new permutation copy of the given one */
PermutationInstance::PermutationInstance(const PermutationInstance &pi) {
	int s = gsl_permutation_size(pi.instance_);
	instance_ = gsl_permutation_alloc(s);
	gsl_permutation_memcpy(instance_, pi.instance_);
	valid_cache_ = pi.valid_cache_;
	fitness_cache_ = pi.fitness_cache_;
}

/* Destroy the PermutationInstance object */
PermutationInstance::~PermutationInstance() {
	gsl_permutation_free(instance_);
}

/* Return the value of permutation at position index */
int PermutationInstance::get(unsigned int index) {
	return (int) gsl_permutation_get(instance_, index);
}

/* Set the value of permutation at position index */
void PermutationInstance::set(unsigned int index, unsigned int value) {
	instance_->data[index] = value; 
	valid_cache_ = false; 
}

/* Return the size of the permutation */
int PermutationInstance::getSize() {
	return (int) gsl_permutation_size(instance_);
}

/* Swap a couple of positions in the permutation. This function invalidate the fitness cache value */
void PermutationInstance::swap(unsigned int idx1, unsigned int idx2) {
	gsl_permutation_swap(instance_, idx1, idx2);
	valid_cache_ = false; 
}

/* Return a new BinaryInstance that is created by appling this permutation to the given BinaryInstance */
BinaryInstance* PermutationInstance::applyPermutation(BinaryInstance* instance) {
	BinaryInstance *bi;
	unsigned int size = instance->getSize();
	
	if(gsl_permutation_size(instance_) != size)
		cerr << "PermutationInstance::applyPermutation Permutation and BinaryInstance have different sizes!" << endl;
	
	bi = new BinaryInstance(size);
	for(unsigned int i=0; i<size; i++) {
		bi->set(i, instance->get(gsl_permutation_get(instance_, i)));
	}
	return bi;
}

/* Set the fitness cache value for this permutation */
void PermutationInstance::setFitnessCache(double fitness) {
	fitness_cache_ = fitness; 
	valid_cache_ = true; 
}

/* Return the fitness cache value of this permutation */
double PermutationInstance::getFitnessCache() {
	return fitness_cache_;
}

/* Return true if the fitness cache value is valid, otherwise return false */
bool PermutationInstance::validCache() {
	return valid_cache_;
}

/* Check if the permutation is consistent */
bool PermutationInstance::isValid() {
	if (gsl_permutation_valid(instance_) == 0)
		return true;
	else 
		return false; 
}

/* Return the internal representation of the permutation */
gsl_permutation* PermutationInstance::getPermutation() {
	return instance_;
}
