
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/

#include <iostream>
#include <fstream>

#include "Solver.h"
#include "path.h"
#include "Statistics.h"
#include "GraphicsTool.h"

/* Solver constructor. It receives as parameters an array of pointers to algorithm objects, 
 * the number of these objects and a file name where record data retrived. */
Solver::Solver(Algorithm **algorithms, unsigned int averageOver, string fileName) :
	_algorithms(algorithms),
	_averageOver(averageOver),
	_fileName(fileName){

}

/* Class destructor. */
Solver::~Solver() {
	for (unsigned int i = 0; i < _averageOver; i++)
		delete _algorithms[i];
	delete[] _algorithms;
}

/* This is a general purpose solver, it can handle a run condition and an averaged run mode separately. 
 * This means that a unique functions can handle every combination of both conditios. The solver simply
 * run an execution of the algorithm for each generation and in case for each average sample, than it
 * perform an aggregation of statistics and save it to a file. */
void Solver::solve(Glib::Mutex *mutex, bool *condition, unsigned int generations) {
	unsigned int i, genCounter = 0, fcalls = 0;
	double avg = 0.0, best = -1.0, worst = 1000.0; 
	ofstream simpleStatFile, bestIndividualFile;
	string simpleStatName(TMP_DATA_PATH), bestIndividualName(TMP_DATA_PATH);
	BinaryInstance *bestIndividual = NULL;
	Statistics *statistics;
	
	simpleStatName.append(_fileName);
	simpleStatFile.open(simpleStatName.c_str());
	if (simpleStatFile.fail()) cerr << "Solver: error while opening data file." << endl;
	
	bestIndividualName.append((char*) "BI_");
	bestIndividualName.append(_fileName);
	bestIndividualFile.open(bestIndividualName.c_str(), ios_base::out | ios_base::binary);
	if (bestIndividualFile.fail()) cerr << "Solver: error while opening best individual file." << endl;
	
	do {	
		avg = 0.0; 
		fcalls = 0;
		best = -1.0; 
		worst = 1000;
		
		/* Performs statistics aggregation */
		for (i = 0; i < _averageOver; i++) {
			statistics = _algorithms[i]->computeStatistics();
			avg += statistics->avgFitness();
			fcalls += statistics->fitnessCalls();
			if (statistics->worstInstance() < worst) worst = statistics->worstInstance();
			if (statistics->bestInstance() > best) {
				best = statistics->bestInstance();
				if (bestIndividual != NULL) delete bestIndividual;
				bestIndividual = new BinaryInstance(* statistics->bestIndividual());
			}
			delete statistics; 
		}
		
		avg = avg / (double) _averageOver; 
		fcalls = (int) (fcalls / _averageOver);
		
		/* Save statistics */
		simpleStatFile << genCounter << "\t" << avg << "\t" << best << "\t" << worst << "\t" << fcalls << endl;
		bestIndividualFile.write(bestIndividual->toCharArray(), bestIndividual->getSize());
		
		simpleStatFile.flush();
		bestIndividualFile.flush();
		/* Run algorithms */
		for (i = 0; i < _averageOver; i++)
			_algorithms[i]->run();
		
	} while ((runCondition(mutex, condition)) && (genCounter++ < generations));
	
	if (simpleStatFile.is_open()) simpleStatFile.close();
	if (bestIndividualFile.is_open()) bestIndividualFile.close();
}

/* Returns the name of the algorithm wrapped by this solver. */
string Solver::getAlgorithmName() {
	return _algorithms[0]->getName();
}

/* Checks the run condition trying to get the lock on the mutex, if not simply return true otherwise
 * return the condition state. It does it in this way in order to avoid time waste. */
bool Solver::runCondition(Glib::Mutex *mutex, bool *condition) {
	bool rval = true;
	
	if (mutex == NULL || condition == NULL) return rval;
	if (mutex->trylock()) {
		rval = *condition;
		mutex->unlock();
	}
	
	return rval;
}
