
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/

#include "MIMIC.h"
#include "Comparator.h"

using namespace std;

/* Class constructor. */
MIMIC::MIMIC(int popSize, int offspring, const gsl_rng* rnd) : EDA(popSize, rnd), _offspring(offspring) {

}

/* Class destructor. */
MIMIC::~MIMIC() {
	for (list<RingOfChain*>::iterator it=_chain.begin(); it!=_chain.end(); it++)
		delete *it;
	_chain.clear();
	for (unsigned int i=0; i<_pop.size(); i++)
		delete _pop[i];
	_pop.clear();
}

/* This function defines the single generation step of the algorithm. */
void MIMIC::run() {
	estimation();
	sampling();
	selection();
}

/* Set the objective function reference and initialize the population. */
void MIMIC::setObjectiveFunction(ObjectiveFunction* of) {
	Algorithm::setObjectiveFunction(of);

	for (unsigned int i=0;i<_popSize;i++) {
		BinaryInstance* instance = new BinaryInstance(objectiveFunctionSize_);
		for (unsigned int j=0;j<objectiveFunctionSize_;j++) {
			instance->set(j,gsl_ran_bernoulli(_rnd, 0.5));
		}
		_pop.push_back(instance);
	}
}

/* Estimation of the bivariate chain model. */
void MIMIC::estimation() {
	double empirical_probability[objectiveFunctionSize_];
	double empirical_entropy[objectiveFunctionSize_];
	list<int> bit_index; 
	vector<int> bits;
	
	/* Clear chain from old rings */
	if(_chain.size() != 0) {
		for (list<RingOfChain*>::iterator it=_chain.begin(); it!=_chain.end(); it++)
			delete *it;
		_chain.clear();
	}
	
	/* Calculating empirical probabilities vector analyzing the population*/
	for (unsigned int i=0; i<objectiveFunctionSize_; i++) {
		int sum_ones = 0;
		for (unsigned int j=0; j<_popSize; j++) {
			if (_pop[j]->get(i) == 1)
				sum_ones++;
		}
		empirical_probability[i] = (double) sum_ones/_popSize;
		bits.push_back(i);
	}
	
	random_shuffle(bits.begin(), bits.end());
	for (unsigned int k=0; k<bits.size(); k++)
		bit_index.push_back(bits[k]);
	bits.clear();
	
	/* Calculating empirical entropy vector and minimal entropy */
	empirical_entropy[0] = entropy(empirical_probability[0]);
	double min = empirical_entropy[0];
	int index = 0;
	for (unsigned int i=1; i<objectiveFunctionSize_; i++) {
		empirical_entropy[i] = entropy(empirical_probability[i]);
		if (empirical_entropy[i] < min) {
			min = empirical_entropy[i];
			index = i;
		}
	}
	
	/* Adding last element of chain - Xn */ 	
	_chain.push_front(new RingOfChain(index, empirical_entropy[index]));
	bit_index.remove(index);
	
	/* Calculating rest of chain - Conditional entropy */
	for (int i = objectiveFunctionSize_ - 1; i > 0; i--) {
		index = *bit_index.begin();
		min = conditionalEntropy(index);
		for (list<int>::iterator j=bit_index.begin(); j!=bit_index.end(); j++) {
			double ce = conditionalEntropy(*j);
			if (ce < min) {
				min = ce;
				index = *j;
			}
		}
		bit_index.remove(index);
		_chain.push_front(new RingOfChain(index, empirical_entropy[index]));
	}
	
	/* This will avoid another unuseful calculation of this probability during sampling */
	_chain.back()->setProbability(empirical_probability[_chain.back()->getIndex()]);
	
	/* --> UNCOMMENT THIS LINE FOR PRINT CHAIN VALUES <--
	cout << "Chain: ";
	list<RingOfChain*>::iterator ring;
	for (ring=_chain.begin(); ring !=_chain.end() ; ring++)
		cout << "[" << (*ring)->getIndex() << "]";
	cout << endl; */

	bit_index.clear(); 
}

/* Computes the conditional entropy. */
double MIMIC::conditionalEntropy(int index) {
	int index_x = _chain.front()->getIndex();
	int index_y = index;
	double prob[4], value = 0;
	int sum[4] = {0,0,0,0};
	
	for (unsigned int i=0; i<_popSize; i++) {
		if ((_pop[i]->get(index_x) == 0) && (_pop[i]->get(index_y) == 0)) sum[0]++;
		else if ((_pop[i]->get(index_x) == 0) && (_pop[i]->get(index_y) == 1)) sum[1]++;
		else if ((_pop[i]->get(index_x) == 1) && (_pop[i]->get(index_y) == 0)) sum[2]++;
		else if ((_pop[i]->get(index_x) == 1) && (_pop[i]->get(index_y) == 1)) sum[3]++;
	}

	for (int i=0; i<4; i++) {
		prob[i] = (double) sum[i]/_popSize;
	}
	
	if(prob[0] != 0) value -= prob[0]*log(prob[0]); else return (double) (DBL_MAX - _chain.front()->getEntropy());
	if(prob[1] != 0) value -= prob[1]*log(prob[1]); else return (double) (DBL_MAX - _chain.front()->getEntropy());
	if(prob[2] != 0) value -= prob[2]*log(prob[2]); else return (double) (DBL_MAX - _chain.front()->getEntropy());
	if(prob[3] != 0) value -= prob[3]*log(prob[3]); else return (double) (DBL_MAX - _chain.front()->getEntropy());
	
	value = (double) (value - _chain.front()->getEntropy());
	return value;
}

/* Computes the information entropy. */
double MIMIC::entropy(double probability) {
	double value = 0;
	
	if (probability == 0.0 || probability == 1)
		return DBL_MAX;
	else {
		value -= probability * log(probability);
		value -= (1-probability) * log(1-probability);
		return value;
	}
}

/* Generate |_popSize| new samples and add it to the population */
void MIMIC::sampling() {
	vector<BinaryInstance*> new_pop;
	
	for (int i=0; i<_offspring; i++) {
		BinaryInstance *instance = new BinaryInstance(objectiveFunctionSize_);
		int bit_value = 0, bit_idx;
		double cond_prob;
		
		bit_idx = _chain.back()->getIndex();
		list<RingOfChain*>::reverse_iterator ring;
		for (ring=_chain.rbegin(); ring !=_chain.rend() ; ++ring) {
			if((*ring)->getIndex() == bit_idx) {
				/*Last element of the chain, and the first to computate */
				bit_value = gsl_ran_bernoulli(_rnd, (*ring)->getProbability());
				instance->set(bit_idx, bit_value);
			} else {
			/*Other elements of the chain */
				cond_prob = conditionalProbability((*ring)->getIndex(), bit_idx, bit_value);
				bit_idx = (*ring)->getIndex();
				bit_value = gsl_ran_bernoulli(_rnd, cond_prob);
				instance->set(bit_idx, bit_value);
			}
		}
		new_pop.push_back(instance);
	}
	_pop.insert(_pop.end(), new_pop.begin(), new_pop.end());
	new_pop.clear();
}

/* Return the conditional probability p(index_target = 1 | index_know = know_value) based on old population */
/* p(target=1 | know=value) = #(target=1 & know=value)/#(know=value) */
double MIMIC::conditionalProbability(int index_target, int index_know, int know_value) {
	int n_know = 0;
	int n_know_target = 0;
	
	for (unsigned int i=0; i<_popSize; i++) {
		if(_pop[i]->get(index_know) == know_value) {
			n_know++;
			if(_pop[i]->get(index_target) == 1)
				n_know_target++;
		}
	}
	
	return (double) n_know_target/n_know;
}

/* Truncation selection of new population maintaining original popSize */
void MIMIC::selection() {
	std::sort(_pop.begin(), _pop.end(), Comparator(this));
	for (unsigned int i=_popSize; i<_pop.size(); i++)
		delete _pop[i];
	_pop.erase(_pop.begin() + _popSize, _pop.end());
	
}

/* Returns the algorithm name. */
string MIMIC::getName() {
	return "MIMIC";
}
