 
/*********************************************************************************
 *  Copyright© 2010 Luigi Malagò (malago@elet.polimi.it)                         *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>. 		 *	
 *********************************************************************************/

#include "UnivariateEDA.h"

/* Class constructor. */
UnivariateEDA::UnivariateEDA(unsigned int popSize, const gsl_rng* rnd) : EDA(popSize, rnd) {

}

/* Class constructor with a given probability vector. */
UnivariateEDA::UnivariateEDA(unsigned int popSize, vector<double> prob, const gsl_rng* rnd) : EDA(popSize, rnd) {
	_prob = prob;
}

/* Class destructor. */
UnivariateEDA::~UnivariateEDA() {
	_prob.clear();
}

/* Set the objective function reference and initialize the population. */
void UnivariateEDA::setObjectiveFunction(ObjectiveFunction* of) {
	unsigned int i; 
	
	Algorithm::setObjectiveFunction(of);
	if (_prob.size() == 0 && objectiveFunctionSize_ != 0) {
		for (i = 0; i < objectiveFunctionSize_; i++) {
			_prob.push_back(0.5);
		}
	}
	sampling();
}

/* Deletes previous population and samples a new one according to probability vector (_prob). */
void UnivariateEDA::sampling() {
	unsigned int i, j;
 	
 	for (i = 0; i < _pop.size(); i++) delete _pop[i];
	_pop.clear();
	
	for (i = 0; i < _popSize; i++) {
		BinaryInstance *instance = new BinaryInstance(_prob.size());
		for (j = 0; j < _prob.size(); j++) {
			instance->set(j, gsl_ran_bernoulli(_rnd, _prob[j]));
		}
		_pop.push_back(instance);
	}
}

/* Truncation selection at 50% of the population. */
void UnivariateEDA::selection() {
	unsigned int i;
	
	std::sort(_pop.begin(),_pop.end(),Comparator(this));
	for (i = _pop.size() / 2; i < _pop.size(); i++) {
		delete _pop[i];
	}
	_pop.erase(_pop.begin() + _pop.size() / 2, _pop.end());
}

/* Graphical visualization of joint probability function. */
void UnivariateEDA::printJointProbalityFunction() {
	cout.setf( ios_base::fixed, ios_base::floatfield);
	cout << "eta = [";
	for (unsigned int j=0; j<_prob.size(); j++) {
		if (_prob[j]>.5)
			cout << "#";
		else
			cout << " ";
	}
	cout.setf(ios_base::dec,ios_base::basefield);
	cout << "]" << endl;
}

/* Graphical visualization of joint probability function. */
void UnivariateEDA::printJointProbalityFunctionDec() {
	cout << "eta = [";
	for (unsigned int j=0; j<_prob.size(); j++) {
		if (_prob[j]>=.9)
			cout << "#";
		else if (_prob[j]<=.1)
			cout << " ";
		else
			cout << ((int)(_prob[j]*10));

	cout << "]";
	cout << " (" << fitness(_pop[0]) << ")" << endl; 
	}
}

/* Returns the current probability vector. */
vector<double> UnivariateEDA::getProbabilityVector() {
	return _prob;
}

/* Set the probability vector. */
void UnivariateEDA::setProbabilityVector(vector<double> prob) {
	_prob = prob;
}
