
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/

#ifndef _EVOPTOOL_H_
#define _EVOPTOOL_H_

#include <sigc++/sigc++.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cblas.h>

#include <ObjectiveFunction.h>
#include <ObjectiveFunctions.h>
#include <Solver.h>
#include <path.h>

#define GRAPH_HEIGHT 450
#define GRAPH_WIDTH 600

using namespace std;

/* This class defines the Evoptool engine, which is the core of this tool. */
 
class Evoptool : public sigc::trackable { 
	/*----------------------------------------------------------------------------------------------------------------*/
	/*						TYPES DEFINITION 						  */
	/*----------------------------------------------------------------------------------------------------------------*/
public: 	
	/* This enumeration defines the tool execution mode. */
	typedef enum {
		GUI,
		COMMAND_LINE
	} ExeMode;
	
	/* This enumeration defines the tool error types. */
	typedef enum {
		UNRECOVERABLE_ERROR,
	} EvoptoolError;
	
	/* This enumeration defines the log types. */
	typedef enum {
		MESSAGE,
		WARNING,
		ERROR
	} LogType;
	
	/* This enumeration defines a set of flags used to specify what kind of statistics have to be computed. */
	typedef enum {
		AVG_GRAPH = 0x001,
		BEST_GRAPH = 0x002,
		WORST_GRAPH = 0x004,
		FCALLS_GRAPH = 0x008,
		BASIC_GRAPHS = 0x00F,  		/* Groups toghether the previous four flags. */
		BEST_AS_IMAGE = 0x010
	} StatisticsFlag;
	
	/* This enumeration defines an identifier for each task function. */
	typedef enum {
        	ALT_BITS,
        	BIN_VAL,
        	F3D_DECEPTIVE,
        	F3D_DECEPTIVE_BIPOLAR,
        	F3D_DECEPTIVE_OVERLAPPING,
        	FOUR_PEAKS,
        	LIEPINS_VOSE_FULLY_DECEPTIVE,
        	ONE_MAX,
        	ONE_ZERO_MAX,
        	QUADRATIC,
        	MAX_SAT,
        	SIX_PEAKS,
        	SUM_VAL,
        	TRAP5,
        	SPIN_GLASS,
        	TEXTURE_RESTORATION
        } Tasks;

	/* This enumeration defines an identifier for each algorithm. */        
        typedef enum {
        	SGA_STD,
        	SGA_BIN_TOURN,
        	SGA_UNIFORM_CROSS,
        	CGA,
        	PBIL,
        	UMDA,
        	MIMIC,
        	PBIL_COMIT,
        	HC_COMIT,
        	UNIVARIATE_DEUM,
        	BIVARIATE_DEUM,
        	ISING_DEUM,
        	SBOA
        } Algorithms;
	
	/* This struct defines a data container for a test execution. */
	struct TestParameters {
		bool exportData;
		char* destinationFolder;
		char* archiveName;
		bool averaged;
		bool perfectSquareSize;
		unsigned int averageSamples;
		unsigned int maxGenerations;
		unsigned int statisticsOptions;
		gsl_rng *rng;
		ObjectiveFunction *task;
		vector<Solver*> solvers;
		Glib::Mutex *mutex;
		bool runCondition;
	};
	
	/* Signals type definition. */
	typedef sigc::signal<void, EvoptoolError> SignalError;
	typedef sigc::signal<void, TestParameters*> SignalInitializeTest;
	typedef sigc::signal<void> SignalRunTest;
	typedef sigc::signal<void> SignalFinalizeTest; 
	typedef sigc::signal<void> SignalTestExecuted;
	typedef sigc::signal<void> SignalStopTest;
	typedef sigc::signal<void, unsigned int, unsigned int, unsigned int> SignalComputeStatistics;
	typedef sigc::signal<void, string> SignalNewLog;
	
	/*----------------------------------------------------------------------------------------------------------------*/
	/*					   MEMBER VARIABLES DEFINITION	 					  */
	/*----------------------------------------------------------------------------------------------------------------*/
protected: 

	const ExeMode _mode;
	TestParameters *_testParameters;
	
	/* Signals. */
	SignalError _signalError;
	SignalInitializeTest _signalInitializeTest;
	SignalRunTest _signalRunTest;
	SignalFinalizeTest _signalFinalizeTest;
	SignalTestExecuted _signalTestExecuted;
	SignalStopTest _signalStopTest;
	SignalComputeStatistics _signalComputeStatistics;
	SignalNewLog _signalNewLog;
	
	/* Supports to statistics computation. */
	char **_bestAsImageFiles;
	unsigned int _bestAsImageFilesNum;
	unsigned int _bestAsImageBlockSize;
	bool _bestAsImageFilesAllocated;
	
	/*----------------------------------------------------------------------------------------------------------------*/
	/*						PUBLIC FUNCTIONS 						  */
	/*----------------------------------------------------------------------------------------------------------------*/
public:
        Evoptool(ExeMode mode);
        virtual ~Evoptool();
        
        /* */
        void init();
        void execute(const char *parametersFile);
        gsl_rng* getRNG();

	/* Signal accessors. */
	SignalError signalError();
	SignalInitializeTest signalInitializeTest();
	SignalRunTest signalRunTest();
	SignalFinalizeTest signalFinalizeTest();
	SignalTestExecuted signalTestExecuted();
	SignalStopTest signalStopTest();
	SignalComputeStatistics signalComputeStatistics();
	SignalNewLog signalNewLog();
	
	/*----------------------------------------------------------------------------------------------------------------*/
	/* 						PRIVATE FUNCTIONS			       			  */
	/*----------------------------------------------------------------------------------------------------------------*/
private:
	
	/* Signal in-class handlers. */
	void signalInitializeTestHandler(TestParameters *parameters);
	void signalRunTestHandler();
	void signalFinalizeTestHandler();	
	void signalStopTestHandler();
	void signalComputeStatisticsHandler(unsigned int options, unsigned int height, unsigned int width);
	void signalErrorHandler(EvoptoolError error);
	
	/* Thread functions. */
	void initializerThread();
	void runnerThread();
	void finalizerThread();
	void stopperThread();
	void solverThread(Solver *solver);
	
	/* Parameters file parser functions. */
	void initTestParameters(TestParameters *params);
	
	/* Tasks and Resolutors management functions. */
	void initializeTaskStructures();
	
	/* General purpose utility functions. */
	void printLog(char *filename, int line, char *message, LogType type);
	bool exportDatas(const char *destinationFolder, const char *archiveName);
	bool cleanTemporanyFiles();
	void deleteTestParameters();
	
	/* Statistics managment function. */
	bool configureStatistics();
	bool configureBasicGraphsStatistics();
	bool configureBestAsImageStatistics();
	bool computeStatistics(unsigned int options = 0, unsigned int height = GRAPH_HEIGHT, unsigned int width = GRAPH_WIDTH);
	bool convertBestToImage(const char **files, int fileNumber, const char *destination);

	/* Gnuplot utility functions. */
	bool setGnuplotData(const char *file, string data);
	bool setGnuplotSize(const char *file, int height, int width);
	bool setGnuplotXRange(const char *file, int range = -1);
	bool callGnuplot(const char *file); 
	
	/* Random number generator utility functions. */
	gsl_rng* allocRNG();
	void freeRNG(gsl_rng *r);
};
#endif
