
/*********************************************************************************
 *  Copyright© 2010 Gabriele Valentini (gabriele.valentini@mail.polimi.it)	 *
 *										 *
 *  This file is part of Evoptool.						 *
 * 										 *
 *  Evptool is free software: you can redistribute it and/or modify              *
 *  it under the terms of the GNU General Public License as published by         *
 *  the Free Software Foundation, either version 3 of the License, or            *
 *  (at your option) any later version.                                          *
 *										 *
 *  Evoptool is distributed in the hope that it will be useful,                  *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of  		 *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 		 *
 *  GNU General Public License for more details.				 *
 *										 *
 *  You should have received a copy of the GNU General Public License		 *
 *  along with Evoptool.  If not, see <http://www.gnu.org/licenses/>.    	 *	
 *********************************************************************************/

#include "XMLParser.h"
#include "Algorithms.h"

/* Parser function. */
bool XMLParser::parse(Evoptool::TestParameters *params, const char *file) {
	short int sectionCounter = 0;
	#ifdef LIBXMLCPP_EXCEPTIONS_ENABLED
	try {
  	#endif
    	
    	/* Open xml file */
    	xmlpp::TextReader reader(file);

	/* Skip content until it find testParameters node */
	do {
		if(!reader.read()) return false;
	} while (reader.get_name().compare("testParameters") != 0);
	
	/* Read sections */
	do {	
		if (!reader.read()) return false;
		
		if (reader.get_name().compare("configuration") == 0) {
			/* Configuration section. */
			if (!parseConfiguration(reader, params)) return false;
								
			do {
				if(!reader.read()) return false;
			} while (reader.get_name().compare("configuration") != 0);	
			sectionCounter++;
		} else if (reader.get_name().compare("task") == 0) {
			/* Task section. */
			if (!parseTask(reader, params)) return false;		
						
			do {
				if(!reader.read()) return false;
			} while (reader.get_name().compare("task") != 0);	
			sectionCounter++;
		} else if (reader.get_name().compare("resolutors") == 0) {
			/* Algorithm section. */
			if (!parseResolutors(reader, params)) return false;
				
			sectionCounter++;
		} else { /* Skip node */ }
	} while (reader.get_name().compare("testParameters") != 0 || sectionCounter != 3);
	
	#ifdef LIBXMLCPP_EXCEPTIONS_ENABLED
	} catch(const exception& e) {
    		cerr << "[XMLParser] Exception caught: " << e.what() << endl;
  		return false;
  	}
	#endif
	
	return true;
}

/* Configuration parser. */
bool XMLParser::parseConfiguration(xmlpp::TextReader &reader, Evoptool::TestParameters *params) {
	Glib::ustring value;
	bool errorFlag = false;
	
	value = readNodeContent(reader, "exportData", &errorFlag);
	if (errorFlag) return false;
	params->exportData = toBool(value);

	if (params->exportData) {
		value = readNodeContent(reader, "destinationFolder", &errorFlag);
		if (errorFlag) return false;
		params->destinationFolder = toChars(value);
			
		value = readNodeContent(reader, "archiveName", &errorFlag);
		if (errorFlag) return false;
		params->archiveName = toChars(value);
	}

	value = readNodeContent(reader, "perfectSquareSize", &errorFlag);
	if (errorFlag) return false;
	params->perfectSquareSize = toBool(value);

	value = readNodeContent(reader, "averageTest", &errorFlag);
	if (errorFlag) return false;
	params->averaged = toBool(value);

	value = readNodeContent(reader, "averageSamples", &errorFlag);
	if (errorFlag) return false;
	params->averageSamples = toInt(value);

	value = readNodeContent(reader, "maxGenerations", &errorFlag);
	if (errorFlag) return false;
	params->maxGenerations = toInt(value);
	
	return true;
}

/* Task section parser. */
bool XMLParser::parseTask(xmlpp::TextReader &reader, Evoptool::TestParameters *params) {
	bool errorFlag = false;
	Glib::ustring value; 
	int task, size, peaks;
	char* file;
			
	value = readNodeContent(reader, "name", &errorFlag);
	if (errorFlag) return false;
	task = toInt(value);

	value = readNodeContent(reader, "size", &errorFlag);
	if (errorFlag) return false;
	size = toInt(value);
	
	if (params->perfectSquareSize) {
		size = pow(floor(sqrt(size)), 2);
	}
			
	switch (task) {
		case Evoptool::ALT_BITS: params->task = new AltBits(size); break;	
		case Evoptool::BIN_VAL: params->task = new BinVal(size); break; 
        	case Evoptool::F3D_DECEPTIVE: params->task = new F3Deceptive(size); break;
        	case Evoptool::F3D_DECEPTIVE_BIPOLAR: params->task = new F3DeceptiveBipolar(size); break;
        	case Evoptool::F3D_DECEPTIVE_OVERLAPPING: 
        		params->task = new F3DeceptiveOverlapping(size); 
        		break;
        	case Evoptool::FOUR_PEAKS: 
        		value = readNodeContent(reader, "peaks", &errorFlag);
			if (errorFlag) return false;
			peaks = toInt(value);
        		params->task = new FourPeaks(size, peaks);
        		break; 
        	case Evoptool::LIEPINS_VOSE_FULLY_DECEPTIVE: 
        		params->task = new LiepinsVoseFullyDeceptive(size); 
       			break;
        	case Evoptool::ONE_MAX: params->task = new OneMax(size); break;
        	case Evoptool::ONE_ZERO_MAX: params->task = new OneZeroMax(size); break;
        	case Evoptool::QUADRATIC: params->task = new Quadratic(size); break;
        	case Evoptool::MAX_SAT: params->task = new SatLib((char*) SATLIB_PATH); break;
        	case Evoptool::SIX_PEAKS: break;
	       		value = readNodeContent(reader, "peaks", &errorFlag);
			if (errorFlag) return false;
			peaks = toInt(value);        			
        		params->task = new SixPeaks(size, peaks);
        		break; 
        	case Evoptool::SUM_VAL: params->task = new SumVal(size); break;
        	case Evoptool::TRAP5: params->task = new Trap5(size); break;
        	case Evoptool::SPIN_GLASS: 
        	 	value = readNodeContent(reader, "filePath", &errorFlag);
		 	if (errorFlag) return false;
			file = toChars(value);
       			params->task = new SpinGlass(file, size);
       			break;
        	case Evoptool::TEXTURE_RESTORATION: 
        	        value = readNodeContent(reader, "filePath", &errorFlag);
			if (errorFlag) return false;
			file = toChars(value);
        		params->task = new TextureRestoration(file);
        		params->statisticsOptions |= Evoptool::BEST_AS_IMAGE;
        		break;
        	default: return false; 
	}
	
	return true;
}

/* Resolutors section parser. */
bool XMLParser::parseResolutors(xmlpp::TextReader &reader, Evoptool::TestParameters *params) {
	  
        AlgorithmParameters algorithmParameters[] = {
        	{ Evoptool::SGA_STD, 2, (char*) "ID" }, 		// Pop, MutRate
        	{ Evoptool::SGA_BIN_TOURN, 2, (char*) "ID" },		// Pop, MutRate
        	{ Evoptool::SGA_UNIFORM_CROSS, 2, (char*) "ID" },	// Pop, MutRate
        	{ Evoptool::CGA, 1, (char*) "I" }, 			// Pop
        	{ Evoptool::PBIL, 2, (char*) "ID" },			// Pop, Gamma
        	{ Evoptool::UMDA, 1, (char*) "I" },			// Pop, Gamma
        	{ Evoptool::MIMIC, 2, (char*) "II" },			// Pop, Samples
        	{ Evoptool::PBIL_COMIT, 7, (char*) "IIIDDII" },		// Pop, Samples, MaxInfluence, Gamma, Mutrate, 
        								// PercSelect, Iteration
        	{ Evoptool::HC_COMIT, 5, (char*) "IIIID" },		// Pop, Samples, MaxInfluence, Patience, Mutrate
        	{ Evoptool::UNIVARIATE_DEUM, 3, (char*) "IDI" },	// Pop, CoolRate, GibbsIterations
        	{ Evoptool::BIVARIATE_DEUM, 3, (char*) "IDI" },		// Pop, CoolRate, GibbsIterations
        	{ Evoptool::ISING_DEUM, 3, (char*) "IDI" },		// Pop, CoolRate, GibbsIterations
        	{ Evoptool::SBOA, 3, (char*) "III" },			// Pop, MaxIncoming, ParentPerc
        };
	
	do {
		if (!reader.read()) return false;
		
		if (reader.get_name().compare("algo") == 0) {
			int i, iC = 0, dC = 0, id, intVal[10];
			double doubleVal[10];
			char *paramType;
			bool errorFlag = false;
			Glib::ustring value; 
			string fileName;
			Algorithm **algos;
			
			if (!reader.has_attributes()) return false;
			reader.move_to_first_attribute();
			id = toInt(reader.get_value());
			
			paramType = algorithmParameters[id].parametersType;
			for (i = 0; i < algorithmParameters[id].parametersCount; i++) {
				if (paramType[i] == 'I') {
					value = readNodeContent(reader, "int", &errorFlag);
		 			if (errorFlag) return false;
					intVal[iC] = toInt(value);
					iC++;
				} else if (paramType[i] == 'D') {
					value = readNodeContent(reader, "double", &errorFlag);
		 			if (errorFlag) return false;
					doubleVal[dC] = toDouble(value);
					dC++;
				} else {
					cerr << "[XMLParser] Parameter type unexpected." << endl;
				}
			}
			
			do {
				if(!reader.read()) return false;
			} while (reader.get_name().compare("algo") != 0);
			
			algos  = new Algorithm*[params->averageSamples];
			
			for (i = 0; i < (int) params->averageSamples; i++) {
				switch (id) {
					case Evoptool::SGA_STD: 
						fileName = "SGA.dat";
        					algos[i] = new SGA(intVal[0], doubleVal[0], params->rng);
						algos[i]->setObjectiveFunction(params->task);
						break;
        				case Evoptool::SGA_BIN_TOURN: 
	        				fileName = "SGABinaryTournment.dat";
        					algos[i] = new SGABinTourn(intVal[0], doubleVal[0], params->rng);
						algos[i]->setObjectiveFunction(params->task);
						break;
        				case Evoptool::SGA_UNIFORM_CROSS: 
	        				fileName = "SGAPopWiseUniformCrossover.dat";
        					algos[i] = new SGAPopWiseCross(intVal[0], doubleVal[0], params->rng);
						algos[i]->setObjectiveFunction(params->task);
				        	break;
        				case Evoptool::CGA: 
        					fileName = "cGA.dat";
        					algos[i] = new cGA(intVal[0], params->rng);
						algos[i]->setObjectiveFunction(params->task);
        					break;
        				case Evoptool::PBIL: 
        					fileName = "PBIL.dat";
						algos[i] = new PBIL(intVal[0], doubleVal[0], params->rng);
						algos[i]->setObjectiveFunction(params->task);
        					break;
        				case Evoptool::UMDA: 
        					fileName = "UMDA.dat";
						algos[i] = new UMDA(intVal[0], params->rng);
						algos[i]->setObjectiveFunction(params->task);
        					break;
        				case Evoptool::MIMIC: 
						fileName = "MIMIC.dat";
						algos[i] = new MIMIC(intVal[0], intVal[1], params->rng);
						algos[i]->setObjectiveFunction(params->task);
              					break; 
        				case Evoptool::PBIL_COMIT: 
	        				fileName = "pbilCOMIT.dat";
						algos[i] = new pbilCOMIT(intVal[0], intVal[1], intVal[2], doubleVal[0], doubleVal[1], intVal[3], intVal[4], params->rng);
						algos[i]->setObjectiveFunction(params->task);
        					break; 
        				case Evoptool::HC_COMIT: 
        					fileName = "hcCOMIT.dat";
						algos[i] = new hcCOMIT(intVal[0], intVal[1], intVal[2], intVal[3], doubleVal[0], params->rng);
						algos[i]->setObjectiveFunction(params->task);
        					break; 
        				case Evoptool::UNIVARIATE_DEUM: 
        					fileName = "UnivariateDEUM.dat";
						algos[i] = new UnivariateDEUM(intVal[0], doubleVal[0], intVal[1], params->rng);
						algos[i]->setObjectiveFunction(params->task);
        					break; 
        				case Evoptool::BIVARIATE_DEUM: 
        					fileName = "BivariateDEUM.dat";
						algos[i] = new BivariateDEUM(intVal[0], doubleVal[0], intVal[1], 1, params->rng);
						algos[i]->setObjectiveFunction(params->task);
						break; 
        				case Evoptool::ISING_DEUM: 
        					fileName = "IsingDEUM.dat";
						algos[i] = new IsingDEUM(intVal[0], doubleVal[0], intVal[1], params->rng);
						algos[i]->setObjectiveFunction(params->task);
						break; 
        				case Evoptool::SBOA: 
						cerr << "[XMLParser] sBOA not supported in this version of the tool." << endl;
						return false;
        					break; 
					default: return false;
				}
			}
			
			params->solvers.push_back(new Solver(algos, params->averageSamples, fileName));
	
		} else { /* Skip Node */ }
	
	} while (reader.get_name().compare("resolutors") != 0);

	return true;
}

/* Reads the content of a composite node (opentag, value, closetag) without children. Skips everything until it 
   find the correct node and its content. Signals errors by setting the errorFlag pointer. */
Glib::ustring XMLParser::readNodeContent(xmlpp::TextReader &reader, const Glib::ustring &nodeName, bool *errorFlag) {
	Glib::ustring value; 
	
	do {
		if(!reader.read()) *errorFlag = true;
	} while (reader.get_name().compare(nodeName) != 0);
	
	if(!reader.read()) *errorFlag = true;
	value = reader.get_value();
	
	if(!reader.read()) *errorFlag = true;
	if(reader.get_name().compare(nodeName) != 0) *errorFlag = true;
	
	return value;
} 

/* Parses the string value as an integer. */
int XMLParser::toInt(Glib::ustring value) {
	int val;
	istringstream(value.raw()) >> val;
	return val;
}

/* Parses the string value as an array of chars. */
char* XMLParser::toChars(Glib::ustring value) {
	char *val = new char[100];
	sprintf(val, "%s", value.c_str());
	return val;
}

/* Parses the string value as a bool. */
bool XMLParser::toBool(Glib::ustring value) {
	if (value.compare("yes") == 0) return true;
	else return false;
}

/* Parses the string value as a double. */
double XMLParser::toDouble(Glib::ustring value) {
	double val; 
	istringstream(value.raw()) >> val;
	return val;
}
